# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


"""And the earth was without form, and void; and darkness was upon the face of the deep.
"""

import os
import sys
import copy
import inspect
import types

from nevow import inevow
from nevow import util
from nevow.compy import MetaInterface, Interface, Adapter, registerAdapter, implements
from nevow import compy


class count(object):
    def __init__(self):
        self.id = 0
    def next(self):
        self.id += 1
        return self.id

nextId = count().next


class InputError(Exception):
    """A Typed instance was unable to coerce from a string to the
    appropriate type.
    """
    def __init__(self, reason):
        self.reason = reason
    
    def __str__(self):
        return self.reason


class ValidateError(Exception):
    """A Binding instance was unable to coerce all it's arguments
    from a dictionary of lists of strings to the appropriate types.
    """
    def __init__(self, errors, formErrorMessage=None, partialForm=None):
        self.errors = errors
        self.formErrorMessage = formErrorMessage
        if partialForm is None:
            self.partialForm = {}
        else:
            self.partialForm = partialForm

    def __str__(self):
        return self.formErrorMessage


class IAutomaticConfigurable(Interface):
    """An object is said to implement IAutomaticConfigurable if
    it __implements__ any TypedInterfaces. When this object
    is configured, discovering binding names, discovering bindings,
    and posting forms along with calling methods in response to
    form posts and setting properties in response to form posts.
    """


class IConfigurable(Interface):
    """An adapter which implements TypedInterfaces for an object
    of the type for which it is registered, provides properties
    which will get and set properties of the adaptee, and methods
    which will perform operations on the adaptee when called.
    
    Web Specific Note: When you implement this interface, you should
    subclass freeform.Configurable instead of implementing directly,
    since it contains convenience functionality which eases implementing
    IConfigurable.
    """
    def getBindingNames(self, context):
        """Return a list of binding names which are the names of all
        the forms which will be rendered for this object when this
        object is configured.
        """

    def getBinding(self, context, name):
        """Return a Binding instance which contains Typed instances
        which describe how to render a form which will gather input
        for the 'name' binding (either a property or a method)
        """

    def postForm(self, context, bindingName, arguments):
        """Handle a form post which configures something about this
        object.
        """

    postLocation = property(doc="""The location of this object in the
    URL. Forms described by bindings returned from getBindingNames
    will be posted to postLocation + '/freeform_post!' + bindingName
    """)

    def __adapt__(other, registry):
        if implements(other, TypedInterface):
            return TypedInterfaceConfigurable(other)


class ITyped(Interface):
    """Typeds correspond roughly to <input> tags in HTML, or
    with a complex type, more than one <input> tag whose input
    is validated and coerced as a unit.
    """
    def coerce(self, val):
        """OPTIONAL.
        
        If this method exists, it will be called to coerce the input
        from a string into a value suitable for the type described
        by the implementor. If coercion fails, coerce should raise
        InputError with a suitable error message to be shown to
        the user.
        
        May return a Deferred.
        """

    def coerceWithBinding(self, val, binding):
        """OPTIONAL.
        
        If this method exists, it will be called to coerce the input
        from a string into a value suitable for the type described
        by the implementor. This method is passed the actual
        object which is being configured, the IConfigurable original,
        as an aid to coercion.
        
        This method is tried first; if it does not exist, coerce is
        tried.
        
        May return a Deferred.
        """

    label = property(doc="""The short label which will describe this
        parameter/properties purpose to the user.""")
    
    description = property(doc="""A long description which further describes the sort
        of input the user is expected to provide.""")
    
    default = property(doc="""A default value that may be used as an initial value in
        the form.""")

    complexType = property(doc="""Whether or not this Typed
        is a "simple" type and the infrastructure should render label,
        description, and error UI automatically, or this type is
        "complex" in which case it will be required to render all UI
        including UI which is normally common to all Typed UI.
        
        This MAY BE DEPRECATED if a better implementation is
        devised.
        """)


class Typed(object):
    """A typed value. Subclasses of Typed are constructed inside of
    TypedInterface class definitions to describe the types of properties,
    the parameter types to method calls, and method return types.
    
    label: The short label which will describe this parameter/proerties
        purpose to the user.
    
    description: A long description which further describes the sort
        of input the user is expected to provide.
    
    default: A default value that may be used as an initial value in
        the form.
    """
    __implements__ = ITyped

    complexType = False

    def __init__(self, label='', description='', default='', allowNone=False, **attributes):
        self.id = nextId()
        self.label = label
        self.description = description
        self.default = default
        self.allowNone=allowNone
        self.attributes = attributes

    def getAttribute(self, name, default=None):
        return self.attributes.get(name, default)

    def coerce(self, val):
        raise NotImplementedError, "Implement in subclass"


class AllowNoneMixin:
    """Mixin for typed types that helps with checking for allowNone."""

    allowNoneFailMessage = 'Please enter a value.'

    def coerce(self, val):
        if val == '':
            if self.allowNone:
                return None
            raise InputError(self.allowNoneFailMessage)
        return val

    

#######################################
## External API; your code will create instances of these objects
#######################################

class String(Typed, AllowNoneMixin):
    """A string that is expected to be reasonably short and contain no
    newlines or tabs.

    strip: remove leading and trailing whitespace.
    """

    allowNoneFailMessage = 'Please enter a string.'

    def __init__(self, *args, **kwargs):
        if 'strip' in kwargs:
            self.strip = kwargs['strip']
            del kwargs['strip']
        else:
            self.strip = False
        Typed.__init__(self, *args, **kwargs)

    def coerce(self, val):
        if self.strip:
            val = val.strip()
        return AllowNoneMixin.coerce(self, val)


class Text(String):
    """A string that is likely to be of a significant length and
    probably contain newlines and tabs.
    """


class Password(String):
    allowNoneFailMessage = 'Please enter a password.'


class FileUpload(Typed, AllowNoneMixin):

    allowNoneFailMessage = 'Please enter a file name.'

    def coerce(self, val):
        AllowNoneMixin.coerce(self, val.filename)
        return val


class Integer(Typed, AllowNoneMixin):

    allowNoneFailMessage = 'Please enter an integer.'

    def coerce(self, val):
        val = AllowNoneMixin.coerce(self, val)
        if val is None:
            return None
        try:
            return int(val)
        except ValueError:
            if sys.version_info < (2,3): # Long/Int aren't integrated
                try:
                    return long(val)
                except ValueError:
                    raise InputError("%r is not an integer." % val)
            
            raise InputError("%r is not an integer." % val)

    
class Real(Typed, AllowNoneMixin):

    allowNoneFailMessage = 'Please enter a real number.'

    def coerce(self, val):
        val = AllowNoneMixin.coerce(self, val)
        if val is None:
            return None
        try:
            return float(val)
        except ValueError:
            raise InputError("%r is not a real number." % val)


class Boolean(Typed):
    def coerce(self, val):
        if self.allowNone and val=='':
            return None
        elif val == 'False':
            return False
        elif val == 'True':
            return True
        raise InputError("%r is not a boolean" % val)


class FixedDigitInteger(Integer):
    
    def __init__(self, digits = 1, *args, **kw):
        Integer.__init__(self, *args, **kw)
        self.digits = digits
        self.allowNoneFailMessage = \
            'Please enter a %d digit integer.' % self.digits

    def coerce(self, val):
        if self.allowNone and val=='':
            return None
        v = Integer.coerce(self, val)
        if len(str(v)) != self.digits:
            raise InputError("Number must be %s digits." % self.digits)
        return v


class Directory(Typed, AllowNoneMixin):
    
    allowNoneFailMessage = 'Please enter a directory name.'

    def coerce(self, val):
        val = AllowNoneMixin.coerce(self, val)
        if val is None:
            return None
        if not os.path.exists(val):
            raise InputError("The directory %r does not exist." % val)
        return val


class Choice(Typed):
    """Allow the user to pick from a list of "choices", or a list of choices found
    by accessing the list in the attribute "choicesAttribute" of the object we
    are configuring. The elements of the list will be rendered by calling the 
    function passed to stringify, which is by default "str".
    """
    def __init__(self, choices=None, choicesAttribute=None, stringify=str, *args, **kw):
        Typed.__init__(self, *args, **kw)
        if choices is None:
            self.choices = []
        else:
            self.choices = choices
        self.choicesAttribute = choicesAttribute
        self.stringify = stringify

    def coerceWithBinding(self, val, binding):
        """Coerce a value with the help of an object, which is the object
        we are configuring.
        """
        if self.allowNone and val == '':
            return None
        try:
            val = int(val)
        except ValueError:
            raise InputError("%r is an invalid choice." % val)
        if self.choicesAttribute is not None:
            choices = getattr(binding, self.choicesAttribute)
        else:
            choices = self.choices
        return choices[val]


class Any(object):
    """Marker which indicates any object type.
    """


class Object(Typed):
    complexType = True
    def __init__(self, interface=Any, *args, **kw):
        Typed.__init__(self, *args, **kw)
        self.interface = interface

    def __repr__(self):
        if self.interface is None:
            return "Object(None)"
        return "Object(interface=%s)" % util.qual(self.interface)


class Group(Object):
    def coerce(self, value):
        print "COERCING GROUP"


class IActionableType(Interface):
    """A type which can have action methods associated with it.
    Currently only List. Probably can be extended to more things.
    """
    def attachActionBindings(self, possibleActions):
        """Attach some MethodBinding instances if they are actions
        for this ActionableType.
        """

    def getActionBindings(self):
        """Return a list of the MethodBinding instances which represent
        actions which may be taken on this ActionableType.
        """


class List(Object):
    __implements__ = IActionableType, Object.__implements__

    complexType = True
    def __init__(self, actions=None, header='', footer='', separator='', *args, **kw):
        """Actions is a list of action methods which may be invoked on one
        or more of the elements of this list. Action methods are defined
        on a TypedInterface and declare that they take one parameter
        of type List. They do not declare themselves to be autocallable
        in the traditional manner. Instead, they are passed in the actions
        list of a list Property to declare that the action may be taken on
        one or more of the list elements.
        """
        if actions is None:
            actions = []
        self.actions = actions
        self.header = header
        self.footer = footer
        self.separator = separator
        Object.__init__(self, *args, **kw)

    def coerce(self, data):
        return data

    def __repr__(self):
        if self.interface is not None:
            return "List(interface=%s)" % util.qual(self.interface)
        return "List()"

    def attachActionBindings(self, possibleActions):
        ## Go through and replace self.actions, which is a list of method
        ## references, with the MethodBinding instance which holds 
        ## metadata about this function.
        act = self.actions
        for method, binding in possibleActions:
            if method in act:
                act[act.index(method)] = binding

    def getActionBindings(self):
        return self.actions


class Dictionary(List):
    pass


class Table(Object):
    complexType = True
    def __repr__(self):
        if self.interface is not None:
            return "Table(interface=%s)" % util.qual(self.interface)
        return "Table()"


class Request(Typed):
    complexType = True ## Don't use the regular form


class Button(Typed):
    def coerce(self, data):
        return data


class Compound(Typed):
    complexType = True
    def __init__(self, elements=None, *args, **kw):
        assert elements, "What is the sound of a Compound type with no elements?"
        self.elements = elements
        Typed.__init__(self, *args, **kw)

    def __len__(self):
        return len(self.elements)

    def coerce(self, data):
        return data


class Method(Typed):
    def __init__(self, returnValue=None, arguments=(), *args, **kw):
        Typed.__init__(self, *args, **kw)
        self.returnValue = returnValue
        self.arguments = arguments


def autocallable(method, action="Call", visible=False, **kw):
    """Describe a method in a TypedInterface as being callable through the
    UI. The "action" paramter will be used to label the action button, or the
    user interface element which performs the method call.
    
    Use this like a method adapter around a method in a TypedInterface:
    
    class IFoo(TypedInterface):
        def doSomething(self):
            '''Do Something
            
            Do some action bla bla'''
            return None
        doSomething = autocallable(doSomething, action="Do it!!")
    """
    method.autocallable = True
    method.id = nextId()
    method.action = action
    method.attributes = kw
    return method


#######################################
## Internal API; formless uses these objects to keep track of
## what names are bound to what types
#######################################


class IBinding(Interface):
    """A Binding corresponds (very) roughly to a <form> tag in HTML.
    A Binding is an object which associates a string name with a Typed
    instance. At the most basic level, Binding instances represent named
    python properties and methods.
    """
    def getArgs(self):
        """Return a copy of this bindings Typed instances; if this binding is a
        Property binding, it will be a list of one Typed istance; if this binding is a
        MethodBinding, it will be a list of all the Typed instances describing
        the method's arguments.
        
        These copies are used during the duration of a form post (initial 
        rendering, form posting, error handling and error correction) to 
        store the user-entered values temporarily in the case of an error
        in one or more input values.
        """

    def getViewName(self):
        """Todo: remove?
        """

    def configure(self, boundTo, results):
        """Configure the object "boundTo" in the manner appropriate
        to this Binding; if this binding represents a property, set the
        property; if this binding represents a method, call the method.
        """

    def coerce(self, val):
        """TODO This is dumb. remove it and leave it on ITyped
        
        Make the code that calls coerce call it on the typed directly
        """

    default = property(doc="""The default value for this binding.""")


class Binding(object):
    """Bindings bind a Typed instance to a name. When TypedInterface is subclassed,
    the metaclass looks through the dict looking for all properties and methods.
    
    If a properties is a Typed instance, a Property Binding is constructed, passing
    the name of the binding and the Typed instance.
    
    If a method has been wrapped with the "autocallable" function adapter,
    a Method Binding is constructed, passing the name of the binding and the
    Typed instance. Then, getargspec is called. For each keyword argument
    in the method definition, an Argument is constructed, passing the name
    of the keyword argument as the binding name, and the value of the
    keyword argument, a Typed instance, as the binding typeValue.
    
    One more thing. When an autocallable method is found, it is called with
    None as the self argument. The return value is passed the the Method
    Binding when it is constructed to keep track of what the method is
    supposed to return.
    """
    __implements__ = IBinding,

    def __init__(self, name, typedValue, id=0):
        self.id = id
        self.name = name
        self.typedValue = ITyped(typedValue)

        # pull these out to remove one level of indirection...
        self.description = typedValue.description
        if typedValue.label:
            self.label = typedValue.label
        else:
            self.label = nameToLabel(name)
        self.default = typedValue.default
        self.complexType = typedValue.complexType

    def __repr__(self):
        return "<%s %s=%s at 0x%x>" % (self.__class__.__name__, self.name, self.typedValue.__class__.__name__, id(self))

    def getArgs(self):
        """Return a *copy* of this Binding.
        """
        return (Binding(self.name, self.original, self.id), )

    def getViewName(self):
        return self.original.__class__.__name__.lower()

    def configure(self, boundTo, results):
        raise NotImplementedError, "Implement in subclass"

    def coerce(self, val):
        if hasattr(self.original, 'coerce'):
            return self.original.coerce(val)
        return val


class Argument(Binding):
    pass


class Property(Binding):
    action = 'Change'
    def configure(self, boundTo, results):
        ## set the property!
        setattr(boundTo, self.name, results[self.name])
        return "Property %r set successfully." % self.name


class MethodBinding(Binding):
    def __init__(self, name, typeValue, id=0, action="Call", attributes = {}):
        Binding.__init__(self, name, typeValue,  id)
        self.action = action
        self.arguments = typeValue.arguments
        self.returnValue = typeValue.returnValue
        self.attributes = attributes

    def getAttribute(self, name):
        return self.attributes.get(name, None)

    def configure(self, boundTo, results):
        bound = getattr(boundTo, self.name)
        rv = bound(**results)
        if rv is None:
            return "Method %r called successfully." % self.name
        return rv

    def getArgs(self):
        """Make sure each form post gets a unique copy of the argument list which it can use to keep
        track of values given in partially-filled forms
        """
        copied = self.typedValue.arguments[:]
        curid = 0
        for n in xrange(len(copied)-1, -1, -1):
            arg=copied[n]
            curid = arg.id
            if isinstance(arg, Button):
                break
        else:
            copied.append(
                Argument(
                    'submit',
                    Button(label=self.getAttribute('label') or 'Submit', description=self.getAttribute('description') or "Submit this form"),
                    curid + 1
                )
            )

        return copied


class ElementBinding(Binding):
    """An ElementBinding binds a key to an element of a container.
    For example, ElementBinding('0', Object()) indicates the 0th element
    of a container of Objects. When this ElementBinding is bound to
    the list [1, 2, 3], resolving the binding will result in the 0th element,
    the object 1.
    """
    pass


class GroupBinding(Binding):
    """A GroupBinding is a way of naming a group of other Bindings.
    The typedValue of a GroupBinding should be a Configurable.
    The Bindings returned from this Configurable (usually a TypedInterface)
    will be rendered such that all fields must/may be filled out, and all
    fields will be changed at once upon form submission.
    """
    def __init__(self, name, typedValue, id=0):
        """Hack to prevent adaption to ITyped while the adapters are still
        being registered, because we know that the typedValue should be
        a Group when we are constructing a GroupBinding.
        """
        self.id = id
        self.name = name
        self.typedValue = Group(typedValue)

        # pull these out to remove one level of indirection...
        self.description = typedValue.description
        if typedValue.label:
            self.label = typedValue.label
        else:
            self.label = nameToLabel(name)
        self.default = typedValue.default
        self.complexType = typedValue.complexType

    def configure(self, boundTo, group):
        print "CONFIGURING GROUP BINDING", boundTo, group


def _sorter(x, y):
    return cmp(x.id, y.id)


class _Marker(object):
    pass


def caps(c):
    return c.upper() == c


def nameToLabel(mname):
    labelList = []
    word = ''
    lastWasUpper = False
    for letter in mname:
        if caps(letter) == lastWasUpper:
            # Continuing a word.
            word += letter
        else:
            # breaking a word OR beginning a word
            if lastWasUpper:
                # could be either
                if len(word) == 1:
                    # keep going
                    word += letter
                else:
                    # acronym
                    # we're processing the lowercase letter after the acronym-then-capital
                    lastWord = word[:-1]
                    firstLetter = word[-1]
                    labelList.append(lastWord)
                    word = firstLetter + letter
            else:
                # definitely breaking: lower to upper
                labelList.append(word)
                word = letter
        lastWasUpper = caps(letter)
    if labelList:
        labelList[0] = labelList[0].capitalize()
    else:
        return mname.capitalize()
    labelList.append(word)
    return ' '.join(labelList)


def labelAndDescriptionFromNameAndDocstring(name, docstring):
    if docstring is None:
        docstring = ''
    docs = filter(lambda x: x, [x.strip() for x in docstring.split('\n')])
    if len(docs) > 1:
        return docs[0], '\n'.join(docs[1:])
    else:
        return nameToLabel(name), '\n'.join(docs)


class MetaTypedInterface(MetaInterface):
    """The metaclass for TypedInterface. When TypedInterface is subclassed,
    this metaclass' __new__ method is invoked. The Typed Binding introspection
    described in the Binding docstring occurs, and when it is all done, there will
    be three attributes on the TypedInterface class:
    
    __methods__: An ordered list of all the MethodBinding instances produced
        by introspecting all autocallable methods on this TypedInterface

    __properties__: An ordered list of all the Property Binding instances produced
        by introspecting all properties which have Typed values on this TypedInterface

    __spec__: An ordered list of all methods and properties
    
    These lists are sorted in the order that the methods and properties appear
    in the TypedInterface definition.
    
    For example:
    
    class Foo(TypedInterface):
        bar = String()
        baz = Integer()
        
        def frotz(self): pass
        frotz = autocallable(frotz)
        
        xyzzy = Float()
        
        def blam(self): pass
        blam = autocallable(blam)

    Once the metaclass __new__ is done, the Foo class instance will have three
    properties, __methods__, __properties__, and __spec__,
    """
    ## Todo try to remove these
    label = "LABEL"
    name = "NAME"
    description = "DESCRIPTION"
    default = "DEFAULT"
    complexType = True
    def __new__(cls, name, bases, dct):
        dct['__id__'] = nextId()
        dct['__methods__'] = methods = []
        dct['__properties__'] = properties = []
        possibleActions = []
        actionAttachers = []
        for key, value in dct.items():
            if key[0] == '_': continue

            if isinstance(value, MetaTypedInterface):
                ## A Nested TypedInterface indicates a GroupBinding
                properties.append(GroupBinding(key, value, value.__id__))
            elif callable(value):
                try:
                    result = value(_Marker)
                except:
                    ## Allow non-autocallable methods in the interface; ignore them
                    continue
                names, _, _, typeList = inspect.getargspec(value)

                if typeList is None:
                    argumentTypes = []
                else:
                    argumentTypes = [
                        Argument(n, argtype, argtype.id) for n, argtype in zip(names[1:], typeList)
                    ]
                label, description = labelAndDescriptionFromNameAndDocstring(key, value.__doc__)

                adapted = ITyped(result, None)
                if adapted is None:
                    adapted = Object(result)
                theMethod = Method(
                    adapted, argumentTypes, label=label, description=description
                )

                if getattr(value, 'autocallable', None):
                    methods.append(
                        MethodBinding(
                            key, theMethod, value.id, value.action, value.attributes))
                else:
                    possibleActions.append((value, MethodBinding(key, theMethod)))
            else:
                if not value.label:
                    value.label = nameToLabel(key)
                if implements(value, IActionableType):
                    actionAttachers.append(value)
                properties.append(
                    Property(key, value, value.id)
                )
        for attacher in actionAttachers:
            attacher.attachActionBindings(possibleActions)
        methods.sort(_sorter)
        properties.sort(_sorter)
        dct['__spec__'] = spec = methods + properties
        spec.sort(_sorter)
        dct['name'] = name
        dct['label'], dct['description'] = labelAndDescriptionFromNameAndDocstring(name, dct.get('__doc__'))
        return type.__new__(cls, name, bases, dct)



class ImproperInputError(Exception):
    """Some malformed input was given to the form machinery;
    usually means a form was posted to a non-existant binding.
    """


class Configurable(object):
    __implements__ = IConfigurable,

    bindingDict = None

    def __init__(self, original):
        self.original = original
        self.boundTo = self

    def getBindingNames(self, context):
        ## Todo: remove this getattr
        ifs = compy.getInterfaces(getattr(self, 'boundTo', self))
        ifs = [
            x for x in ifs if x is not IConfigurable and x is not TypedInterface
        ]
        bindingNames = []
        self.bindingDict = bindingDict = {}
        for interface in ifs:
            ## TypedInterfaces have a __spec__ attribute which is a list of all Typed properties and
            ## autocallable methods
            for binding in getattr(interface, '__spec__', []):
                bindingDict[binding.name] = binding
                if binding.name not in bindingNames:
                    bindingNames.append(binding.name)
                if compy.implements(binding.typedValue, IActionableType):
                    acts = binding.typedValue.actions
                    if acts is None:
                        acts = []
                    for action in acts:
                        bindingDict[action.name] = action
        return bindingNames

    def getDefault(self, forBinding):
        name = forBinding.name
        if hasattr(self, name):
            return getattr(self, name)
        ## Todo: Only do this in ConfigurableAdapter instead of Configurable
        if hasattr(self.boundTo, name):
            return getattr(self.boundTo, name)
        if self.original is not self.boundTo and hasattr(self.original, name):
            return getattr(self.original, name)
        return forBinding.default

    def getBinding(self, context, name):
        if self.bindingDict is None:
            self.getBindingNames(context)
        if self.bindingDict is None:
            self.bindingDict = {}
        binding = getattr(self, 'bind_%s' % name, getattr(self.boundTo, 'bind_%s' % name, None))
        if binding is not None:
            binding = binding(context)
        else:
            try:
                binding = self.bindingDict[name]
            except KeyError:
                raise ImproperInputError, name
        binding.boundTo = self.boundTo
        return binding

    def postForm(self, ctx, bindingName, args):
        """Accept a form post to the given bindingName. The bindingName
        can be dotted to indicate an attribute of this Configurable, eg
        addresses.0.changeEmail. The post arguments are given in args.
        Return a Resource which will be rendered in response.
        """
        from nevow.tags import invisible
        request = ctx.locate(inevow.IRequest)
        pathSegs = bindingName.split('.')
        configurable = self

        cf = ctx.locate(inevow.IConfigurableFactory)
        ## Get the first binding
        firstSeg = pathSegs.pop(0)
        binding = configurable.getBinding(ctx, firstSeg)
        ctx.remember(binding, IBinding)
        ctx.remember(configurable, IConfigurable)
        ## I don't think this works right now, it needs to be fixed.
        ## Most cases it won't be triggered, because we're just traversing a
        ## single binding name
        for seg in pathSegs:
            assert 1 == 0, "Sorry, this doesn't work right now"
            binding = configurable.getBinding(ctx, seg)
            child = self.boundTo
            if not isinstance(binding, GroupBinding):
                accessor = inevow.IContainer(configurable.boundTo, None)
                if accessor is None:
                    child = getattr(configurable.boundTo, binding.name)
                else:
                    child = accessor.child(ctx, binding.name)
            ## If it's a groupbinding, we don't do anything at all for this path segment
            
            ## This won't work right now. We need to push the previous configurable
            ## as the configurableFactory somehow and ask that for hte next binding
            configurable = cf.locateConfigurable(ctx, child)
            ctx = ctx.with(invisible(key=seg))
            ctx.remember(binding, IBinding)
            ctx.remember(configurable, IConfigurable)

        bindingValidator = inevow.IInputValidator(binding)
        return bindingValidator.validate(ctx, binding.boundTo, args)

    def summary(self):
        return "An instance of %s" % self.__class__.__name__

    postLocation = None


class NotFoundConfigurable(Configurable):
    def getBinding(self, context, name):
        raise ImproperInputError, self.original


class TypedInterfaceConfigurable(Configurable):
    def __init__(self, original):
        self.original = original
        self.boundTo = original

    def summary(self):
        return "An instance of %s" % self.original.__class__.__name__

    def __repr__(self):
        return "TypedInterfaceConfigurable(%r)" % self.original


class ListConfigurable(TypedInterfaceConfigurable):
    def getBinding(self, context, name):
        eb = ElementBinding(name, Object())
        eb.boundTo = self.original
        return eb


class GroupConfigurable(TypedInterfaceConfigurable):
    def __init__(self, original, groupInterface):
        TypedInterfaceConfigurable.__init__(self, original)
        self.groupInterface = groupInterface

    bindingDict = None

    def getBindingNames(self, context):
        bindingNames = []
        self.bindingDict = bindingDict = {}
        interface = self.groupInterface
        for binding in getattr(interface, '__spec__', []):
            bindingDict[binding.name] = binding
            if binding.name not in bindingNames:
                bindingNames.append(binding.name)
            if compy.implements(binding.typedValue, IActionableType):
                acts = binding.typedValue.actions
                if acts is None:
                    acts = []
                for action in acts:
                    bindingDict[action.name] = action
        return bindingNames


#######################################
## External API; subclass this to create a TypedInterface
#######################################


class TypedInterface(Interface):
    __metaclass__ = MetaTypedInterface

    """Inherit from this to create interfaces which annotate the types
    of properties and types of parameters methods take and types
    of objects methods return.
    """


