# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


from __future__ import generators

from twisted.internet.defer import Deferred, DeferredList
from twisted.web import resource
from twisted.web import static
from twisted.web import server

from nevow import formless
from nevow import inevow
from nevow.renderer import flatten
from nevow.tags import *
from nevow.inevow import IRendererFactory, IInputValidator
from nevow import util
from nevow import compy
from nevow import url


try:
    enumerate
except:
    def enumerate(collection):
        i = 0
        it = iter(collection)
        while 1:
            yield (i, it.next())
            i += 1


class PrefixerDict(dict):
    def __init__(self, prefix, errors):
        if prefix is None: prefix = ''
        self.prefix = prefix
        self.errors = errors
        dict.__init__(self)

    def __setitem__(self, key, value):
        if key is None:
            key = ''
        if key == '':
            pfxkey = self.prefix
        else:
            pfxkey = '.'.join((self.prefix, key))
        self.errors[pfxkey] = value

    def __getitem__(self, key):
        if key == '':
            pfxkey = self.prefix
        else:
            pfxkey = '.'.join((self.prefix, key))
        return self.errors[pfxkey]

    def update(self, other):
        for key, value in other.items():
            self[key] = value


class IConfigurableKey(compy.Interface):
    """The key of the configurable which is being rendered
    """


class IFormDefaults(compy.Interface):
    """Default values for the current form
    """


class FormDefaults(compy.Adapter):
    def __init__(self, original):
        """original should be a request"""
        self.original = original
        self.defaults = {}

    def setDefault(self, key, value, context=None):
        self.defaults[key] = value

    def getDefault(self, key, context=None):
        #print "getting default for key", key, self.defaults
        # 1) Check on the request
        current = self.defaults.get(key, None)
        if current is None:
            # 2) Check on the session
            if context is not None:
                sessionDefaults = IFormDefaults(context.locate(inevow.IRequest).getSession())
                if sessionDefaults is not self:
                    current = sessionDefaults.getDefault(key)
                    if current is not None:
                        return current
                # 3) Ask the Binding instance for the default values
                try:
                    configurable = context.locate(formless.IConfigurable)
                except KeyError:
                    return ''
                return configurable.getDefault(context.locate(inevow.IData))
        return current

    def getAllDefaults(self, key):
        return PrefixerDict(key, self.defaults)

    def clearAll(self):
        self.defaults = {}


class IFormErrors(compy.Interface):
    """An object which keeps track of which forms have which errors
    """


class FormErrors(compy.Adapter):
    """An object which keeps track of which forms have which errors
    """
    __implements__ = IFormErrors,
    def __init__(self, original):
        """original should be a session"""
        self.original = original
        self.errors = {}

    def setError(self, errorKey, error):
        self.errors[errorKey] = error

    def getError(self, errorKey):
        #print "get error", errorKey, self.__dict__
        return self.errors.get(errorKey)

    def getAllErrors(self, formName):
        return PrefixerDict(formName, self.errors)

    def updateErrors(self, formName, errors):
        PrefixerDict(formName, self.errors).update(errors)

    def clearErrors(self, formName):
        for key in self.errors.keys():
            if key.startswith(formName):
                del self.errors[key]

    def clearAll(self):
        self.errors = {}


class IBindingRenderer(compy.Interface):
    """Something which can render a formless Binding.

    A Binding represents an atomic form which can be
    submitted to cause a named thing to occur;
    a MethodBinding will cause a method to be called;
    a PropertyBinding will cause a property to change.
    """


class IActionRenderer(compy.Interface):
    """An alternate rendering of a formless Binding which
    is usually represented as a smaller, no-input-fields
    toolbar-style button. Should call a MethodBinding.

    An action is distinct from a MethodBinding in that
    an action does not explicitly solicit input from the user,
    but instead gathers all the information it needs to run
    a method from implicit context, such as the render
    context, or the current state of the selection.
    """


class ITypedRenderer(compy.Interface):
    """Something which can render a formless Typed.
    Renders input fields in html which will gather information
    from the user which will be passed to the Typed.coerce
    method when the entire form is submitted.
    """


class IRedirectAfterPost(compy.Interface):
    """A marker interface which can be set to something which can be cast
    to a string to indicate that the browser should be redirected to the
    resultant url after posting the form successfully. This component can
    be set by any form post validators, or by the configurable method which
    is being automatically called if it has access to the request.

    Set this using the following:

    request.setComponent(IRedirectAfterPost, "http://example.com/")
    """


def defaultBindingRenderer(context, data):
    return invisible(render=StringRenderer(data))


def calculatePostLocation(context, data):
    prepath = context.locate(inevow.IRequest).prepath
    if prepath[-1] == '':
        return '.'
    else:
        return prepath[-1]
    return postLocation


def calculatePostURL(context, data):
    postLocation = calculatePostLocation(context, data)
    try:
        configurableKey = context.locate(IConfigurableKey)
    except KeyError:
        #print "IConfigurableKey was not remembered when calculating full binding name for %s in node %s" % (configurable, context.key)
        configurableKey = ''
    if configurableKey is None:
        configurableKey = ''
    bindingName = context.key
    if bindingName is None:
        bindingName = ''
    return "%s/freeform_post!%s!%s" % (postLocation, configurableKey, bindingName)


def keyToXMLID(key):
    """Convert a key into an XML-style ID """
    if not key:
        print 'keyToXMLID: no key, but why?'
        return None
    return '-'.join(key.split('.'))


class BaseInputRenderer(compy.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    def __call__(self, context, data):
        request = context.locate(inevow.IRequest)
        defaults = IFormDefaults(request)
        #print "CONTEXT KEY", context.key, context
        value = defaults.getDefault(context.key, context)
        #print "DEFAULT VALUE", `value`, defaults.__dict__

        if isinstance(value, Deferred):
            value.addBoth(self._cb_call, context, data)
            return value
        else:
            return self._cb_call(value, context, data)

    def _cb_call(self, value, context, data):
        request = context.locate(inevow.IRequest)
        errors = IFormErrors(request.getSession())
        errormessage = errors.getError(context.key)
        if not errormessage:
            errormessage = comment["\nNo error for error key: %s\n" % context.key]

        if data.typedValue.getAttribute('compact'):
            slot = invisible()
        else:
            slot = span(_class="freeform-input")
        if value != '' and request.args.has_key(data.name):
            value = request.args[data.name][0]
        if not data.typedValue.getAttribute('immutable'):
            theslotfilled = self.input(context, slot, data, data.name, value)
            if theslotfilled is None:
                ## Breakout; we don't have any options to give the user, so don't render anything
                return ''
        else:
            theslotfilled = slot[value]

        if data.typedValue.getAttribute('hidden'):
            return theslotfilled
        if data.typedValue.getAttribute('compact'):
            return theslotfilled

        return [
            label(_class="freeform-label", _for=keyToXMLID(context.key))[ data.label ],
            theslotfilled, # freeform-input
            span(_class="freeform-error")[ errormessage ],
            div(_class="freeform-description")[ data.description ],
        ]

    def input(self, context, slot, data, name, value):
        raise NotImplementedError, "Implement in subclass"


class PasswordRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        return [
            input(id=keyToXMLID(context.key), name=name, type="password"), " Again ", input(name="%s____2" % name, type="password"),
        ]


class StringRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        if data.typedValue.getAttribute('hidden'):
            T="hidden"
        else:
            T="text"
        return slot[
            input(id=keyToXMLID(context.key), type=T, name=name, value=value)
        ]


class TextRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        return slot[
            textarea(id=keyToXMLID(context.key), name=name, _class="freeform-textarea")[
                value or ''
            ]
        ]


class BooleanRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        ## The only difference here is the "checked" attribute; the value is still the same because
        ## we want true to be passed to the server when the checkbox is checked and the form
        ## is posted.
        if value:
            slot[input(id=keyToXMLID(context.key), type="checkbox", checked="checked", name=name, value='True')]
        else:
            slot[input(id=keyToXMLID(context.key), type="checkbox", name=name, value='True')]

        # HTML forms are so weak. If the checkbox is not checked, no value at all will be
        # in request.args with the name data.name. So let's force the value False to always
        # be in request.args[data.name]. If the checkbox is checked, the value True will
        # be first, and we will find that.
        slot[input(type="hidden", name=name, value="False")]
        return slot

class FileUploadRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        return slot[input(id=keyToXMLID(context.key), type="file", name=name)]


class ChoiceRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        tv = data.typedValue
        if tv.choicesAttribute:
            choices = getattr(context.locate(formless.IConfigurable).boundTo, tv.choicesAttribute)
        else:
            choices = tv.choices

        numChoices = len(choices)
        if numChoices == 0:
            return None

        selecter = select(id=keyToXMLID(context.key), name=name)
        stringify = tv.stringify

        for index, val in enumerate(choices):
            if val == value:
                selecter[option(value=str(index), selected="selected")[stringify(val)]]
            else:
                selecter[option(value=str(index))[stringify(val)]]
        return slot[selecter]


def isGrouped(context, depth):
    """Return True if the IBinding at the given depth is a Group binding.
    """
    try:
        bnd = context.locate(formless.IBinding, depth=depth)
    except KeyError:
        return False
    if isinstance(bnd, formless.GroupBinding):
        return True
    return False


class ButtonRenderer(compy.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    def __call__(self, context, data):
        if isGrouped(context, 3):
            return ''
        return input(type="submit", value=data.label)


class ObjectRenderer(compy.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    def __call__(self, context, data):
        configurable = context.locate(formless.IConfigurable)
        return div(data=getattr(configurable, data.name))[ configure ]


class RequestRenderer(compy.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    def __call__(self, context, data):
        return ''


## Hoorj javascript
LOCATE_AND_CHANGE_CURRENT_SELECTION_INPUT = xml('')
("var sel = document.getElementById('freeform-action-selection!' + this.name);"\
"if (this.checked) { "\
"sel.value += this.value+'-';"\
"} else {"\
" split = sel.value.split('-'); "\
" sel.value = '';"\
" for(var i = 0; i<split.length; i++) {"\
"   if (split[i] && split[i] != this.value) { sel.value +=split[i]+'-' }"\
" }"\
"};")

## This won't actually work outside of quotient right now, sorry
GATHER_CHECKED_VALUES = xml("""updateSelection(this.id)""")


class ListRenderer(compy.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    complexType = True
    styleClass = 'freeform-list'

    def items(self, context, data):
        request = context.locate(inevow.IRequest)
        defaults = IFormDefaults(request)
        itemList = list(defaults.getDefault(context.key, context))
        configurableList = [
            locateIfPossible(context, item) for item in itemList
        ]
        postLocations = [None] * len(configurableList)
        for i, config in enumerate(configurableList):
            postLocations[i] = str(i)
        return zip(postLocations, itemList, configurableList)

    def __call__(self, context, data):
        errors = IFormErrors(context.locate(inevow.IRequest).getSession())
        if isinstance(data, formless.Property):
            ## List properties are implemented
            actions = data.typedValue.actions or []
            bindingName = context.key
            if bindingName is None: bindingName = ''
            if actions:
                postLocation = calculatePostLocation(context, data)
                #topConf = context.locate(formless.IConfigurable)
                configurableKey = context.locate(IConfigurableKey)
                actionForm = form(
                    _class="freeform-action-group",
                    enctype="multipart/form-data",
                    #id="%s!%s" % (configurableKey, bindingName),
                    id="%s" % (bindingName, ),
                    action="%s/freeform-action-post!%s!%s" % (postLocation, configurableKey, bindingName),
                    method="POST",
                    onsubmit=GATHER_CHECKED_VALUES,
                )[
                    input(type="hidden", id="freeform-action-selection!%s!%s" % (configurableKey, bindingName), name="freeform-action-selection", value=""),
                    [invisible(_class="freeform-action", data=act, render=IActionRenderer(act, persist=False)) for act in actions]
                ]
            else:
                actionForm = ''

            items = self.items(context, data)
            if not items:
                
                content = div(_class=self.styleClass) [ data.typedValue.getAttribute('empty', "An empty list.") ]
            else:
                listItems = [None] * len(items)

                for i, (key, value, conf) in enumerate(items):
                    binding = formless.ElementBinding(str(i), formless.Object())
                    d = div(
                        _class="freeform-list-item",
                        key=key,
                        remember=binding
                    )
                    listItems[i] = d
                    confFound = not isinstance(conf, formless.NotFoundConfigurable)
                    ## remove this; this is for the pool browser being embedded in freeform.
                    ## really, you want items embedded in freeform to automatically have
                    ## checkboxes.
                    if confFound and actions:
                        ## data.name makes it possible for the script to calculate 'freeform-action-selection:%s' % data.name
                        ## which is the id of the node which will keep track of which of these checkboxes are selected.
                        d[
                            input(
                                _class="freeform-list-item-checkbox",
                                type="checkbox",
                                name = data.name,
                                value = key,
                                id = "%s!%s!%s" % (configurableKey, bindingName, i))]
                    ctx = context.with(d)
                    ctx.remember(binding, formless.IBinding)
                    ctx.remember(conf, formless.IConfigurable)
                    fake = formless.ElementBinding('', formless.Object())
                    visibility = 'block'
                    if confFound:
                        d[
                            span(_class="freeform-form-label") [
                                ## TODO: Remove getattr by making sure every IConfigurable has a summary method
                                ## TODO: remove this by using ISummary instead
                                getattr(value, 'summary', lambda: str(value))()
                            ],
                            div(
                                _class="binding-content",
                                style="display: %s;" % visibility,
                                data=value
                            )[
                                configure
                            ]
                        ]
                    else:
                        ## No configurable, let's just throw whatever it was in the dom
                        if renderer.getSerializer(value):
                            try:
                                for x in range(100):
                                    theData = context.locate(inevow.IData, x+1)
                                    if not isinstance(theData, formless.Binding):
                                        break
                            except KeyError:
                                print "keyeerror"
                            d[
                                invisible(data=theData)[ value ]
                            ]
                        else:
                            d[
                                "An unconfigurable, unserializable object: %r" % value
                            ]

                content = div(_class=self.styleClass) [ actionForm, data.typedValue.header, listItems, data.typedValue.footer ]
            return content
        else:
            ## List arguments aren't implemented yet
            return 'List argument input forms are not implemented yet.'


class DictionaryRenderer(ListRenderer):
    complexType = True
    styleClass = 'freeform-dictionary'

    __implements__ = IRendererFactory,
    def items(self, context, data):
        request = context.locate(inevow.IRequest)
        defaults = IFormDefaults(request)
        theDict = defaults.getDefault(context.key, context)
        items = theDict.items()
        return [
            (x[0], "%s: %s" % (x[0], x[1]), locateIfPossible(context, x)) for x in items
        ]


def getError(context):
    errors = IFormErrors(context.locate(inevow.IRequest).getSession())
    err = errors.getError(context.key)
    if err is not None:
        return div(_class="freeform-error")[ err ]
    return ''


class GroupBindingRenderer(compy.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, data):
        context.remember(data, formless.IBinding)

        configurable = formless.GroupConfigurable(data.boundTo, data.typedValue.interface)
        context.remember(configurable, formless.IConfigurable)

        bindingNames = configurable.getBindingNames(context)
        bindings = []

        for name in bindingNames:
            bnd = configurable.getBinding(context, name)
            bindings.append(
                invisible(
                    data = bnd,
                    render = IBindingRenderer(bnd, defaultBindingRenderer, persist=False),
                    key=name
                )
            )

        return getError(context), form(id=keyToXMLID(context.key), _class="freeform-group-binding", enctype="multipart/form-data", action=calculatePostURL(context, data), method="POST")[
            bindings,
            input(type="submit")
        ]


class PropertyBindingRenderer(compy.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, data):
        context.remember(data, formless.IBinding)
        typedRenderer = ITypedRenderer(data.typedValue, defaultBindingRenderer, persist=False)
        if getattr(typedRenderer, 'complexType', False):
            return invisible(data=data, render=typedRenderer)

        grouped = isGrouped(context, 2)

        if grouped:
            flat = flatten(renderer.serialize(invisible(data=data, render=typedRenderer, key=data.name), context))
        else:
            flat = flatten(renderer.serialize(invisible(key=data.name)[invisible(data=data, render=typedRenderer)], context))
        if not flat:
            return ''
        if grouped:
            frm = div(_class="freeform-grouped-property-binding")
            subm = ''
        else:
            ## If we're not inside of a GroupBinding, we need a form and a submit button
            frm = form(id=keyToXMLID(context.key), action=calculatePostURL(context, data), _class="freeform-property-binding", method="POST")
            if not data.typedValue.getAttribute('immutable'):
                subm = input(type="submit", name="change", value="Change")
            else:
                subm = ''

        return getError(context), frm[
            flat,
            subm
        ]


class MethodBindingRenderer(compy.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, data):
        if data.getAttribute('invisible'):
            return ''

        context.remember(data, formless.IBinding)
        args = data.getArgs()

        request = context.locate(inevow.IRequest)

        if isGrouped(context, 2):
            frm = div(_class="freeform-grouped-method-binding")
        else:
            ## If we're not inside of a GroupBinding, we need a form
            frm = form(
                id=keyToXMLID(context.key),
                method="POST",
                enctype="multipart/form-data",
                action=calculatePostURL(context, data),
                _class="freeform-method-binding")

        if data.getAttribute('compact'):
            return frm(_class="freeform-compact-method")[[
                span(key=argument.name, data=argument, render=ITypedRenderer(argument.typedValue, defaultBindingRenderer, persist=False))
                for argument in args
            ]]
        return getError(context), frm [
            div(_class="freeform-form-label") [
                data.label,
            ],
            fieldset(_class="binding-content")[
                data.description,
                [
                    div[invisible(key=argument.name, data = argument, render = ITypedRenderer(argument.typedValue, defaultBindingRenderer, persist = False))]
                    for argument in args
                ],
            ],
        ]


class ArgumentBindingRenderer(compy.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, argument):
        context.remember(data, formless.IBinding)
        ## if need be add an extra div here with class freeform-argument-binding
        return div(_class="freeform-argument-binding")[
            invisible(render=ITypedRenderer(argument.typedValue, defaultBindingRenderer, persist=False))
        ]


class ActionRenderer(compy.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, action):
        submit = input(type="submit", name="freeform-action", value=action.name)
        args = action.getArgs()
        ## The first argument is the list argument; this will be taken care of by the checkboxes and some other logic in the page
        return [
            div(key=argument.name, data = argument, render = ITypedRenderer(argument.typedValue, defaultBindingRenderer, persist = False))
            for argument in args[1:-1]
        ], submit

def clearErrors(context, data):
    errs= IFormErrors(context.locate(inevow.IRequest).getSession())
    #errs.errors = {}
    return ''

 
def action(name):
    """Return the form action string for a given binding name, either
    a property or a method of the current configurable
    
    usage:
    form(render=action('foo'))
    """
    def render(ctx, data):
        if ctx.key is None: configurableKey = ''
        else: configurableKey = ctx.key
        ctx.remember(configurableKey, IConfigurableKey)
        ctx.key = name
        configurable = locateConfigurable(ctx, data)
        ctx.remember(configurable, formless.IConfigurable)
        binding = configurable.getBinding(ctx, name)
        ctx.remember(binding, formless.IBinding)
        return ctx.tag(
            data=binding,
            action=calculatePostURL(ctx, data),
            id=ctx.key,
            enctype="multipart/form-data",
            method="POST")
    return render


def argument(name):
    """Render a field for a given argument binding name. The method binding
    to which this argument binding refers is determined by the current context
    key. Eg if context.key is "postForm" and the name passed is 'firstName',
    the renderer will render an input field for the firstName argument of the
    postForm method.
    
    This should be used in conjunction with the action renderer on a form tag,
    to ensure the proper context and post location are set up:
    
    form(render=freeform.action('sendMail'))[
        freeform.argument('to'),
        freeform.argument('doit') ## render the submit button. xxx don't call it 'doit'
    ]
    """
    def render(ctx, data):
        args = data.getArgs()
        for argument in args:
            if argument.name == name:
                break
        else:
            raise RuntimeError("The argument %r was not found in the binding %r." % (name, data.name))
        
        return div(key=argument.name, data = argument, render = ITypedRenderer(argument.typedValue, defaultBindingRenderer, persist = False))
    return render


def binding(name):
    """Render a form for a given binding name, either a property or a method.
    This will render the entire form. No additional skin is necessary.
    """
    def render(ctx, data):
        configurable = locateConfigurable(ctx, data)
        ctx.remember(configurable, formless.IConfigurable)
        binding = configurable.getBinding(ctx, name)
        ## keep track of what it was here
        ctx.remember(ctx.key, IConfigurableKey)
        return invisible(data=binding,
            render=IBindingRenderer(binding),
            key=name)
    return render


def value(name):
    """Render the current value of the binding name as a string.
    The method binding to which this argument binding refers is determined 
    by the current context key. Eg if context.key is "postForm" and the name 
    passed is 'firstName', the renderer will return the current value of the firstName
    argument to the postForm method.
    """
    def render(ctx, data):
        defaults = IFormDefaults(ctx.locate(inevow.IRequest))

        ## This sucks. Refactor configurable.getDefault
        args = data.getArgs()
        for argument in args:
            if argument.name == name:
                break
        else:
            raise RuntimeError("The argument %r was not found in the binding %r." % (name, data.name))
        C = ctx.with(invisible(data=argument))

        D = defaults.getDefault('.'.join([C.key, name]), C)
        if D is None: return ''
        return D
    return render


def error(name):
    """Render the current error message of the binding name as a string.
    The method binding to which this argument binding refers is determined 
    by the current context key. Eg if context.key is "postForm" and the name 
    passed is 'firstName', the renderer will return the current value of the firstName
    argument to the postForm method.
    """
    def render(ctx, data):
        errors = IFormErrors(ctx.locate(inevow.IRequest).getSession())
        fullName = '.'.join([ctx.key, name])
        E = errors.getError(fullName)
        if not E: return comment["\nNo error for error key: %s\n" % fullName]
        return E
    return render


def createConfigurer(message):
    def locateConfigurable(context, data):
        if compy.implements(data, formless.IConfigurable):
            return data
        configurable = None
        ## Only try the resource itself if we haven't already been configuring it.
        ## remove these "check the IResource for IConfigurable" hacks if possible
        try:
            existing = context.locate(formless.IConfigurable)
            if getattr(existing, 'original', None) is data:
                return existing
        except KeyError:
            ## First try the resource itself (hacky)
            res = context.locate(inevow.IResource)
            configurable = formless.IConfigurable(res, None)
        if configurable is None:
            configurable = formless.IConfigurable(data, None)
            if configurable is None:
                return formless.NotFoundConfigurable(message % {'data': data})

        return configurable

    def configure(context, data):
        configurable = locateConfigurable(context, data)
        context.remember(configurable, formless.IConfigurable)

        bindingNames = configurable.getBindingNames(context)
        bindings = []

        for name in bindingNames:
            bnd = configurable.getBinding(context, name)
            bindings.append(
                invisible(
                    data = bnd,
                    render = IBindingRenderer(bnd, defaultBindingRenderer, persist=False),
                    key = name
                )
            )

        return bindings
    return locateConfigurable, configure


locateConfigurable, configure = createConfigurer("An unconfigurable object (%(data)r)")
locateIfPossible, configureIfPossible = createConfigurer("")


def renderForms(configurableKey='', bindingNames=None):
    """Render forms for either the named configurable. If no configurableKey is given,
    the main configurable will be used. If no bindingNames are given, forms will be
    rendered for all bindings described by the configurable.
    """
    def renderer(context, data):

        # Remember the key for when the form post URL is generated.
        context.remember(configurableKey, IConfigurableKey)

        cf = context.locate(inevow.IConfigurableFactory)
        configurable = cf.locateConfigurable(context, configurableKey)
        context.remember(configurable, formless.IConfigurable)

        available = configurable.getBindingNames(context)
        bindings = []

        for name in available:
            if bindingNames is not None and name not in bindingNames:
                continue
            bnd = configurable.getBinding(context, name)
            bindings.append(
                invisible(
                    data = bnd,
                    render = IBindingRenderer(bnd, defaultBindingRenderer, persist=False),
                    key = name
                    )
                )

        return bindings
    
    return renderer


faketag = html()


def exceptblock(f, handler, exception, *a, **kw):
    try:
        result = f(*a, **kw)
    except exception, e:
        return handler(e)
    if isinstance(result, Deferred):
        def _(fail):
            fail.trap(exception)
            return handler(fail.value)
        return result.addErrback(_)
    else:
        return result

    
class GroupBindingValidator(compy.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data):
        ## THE SPEC: self.original.typedValue.interface.__spec__
        spec = self.original.typedValue.interface.__spec__
        resultList = [None] * len(spec)
        message = ''
        results = {}
        failures = {}
        waiters = []
        for i, sub in enumerate(spec):
            def _validate():
                # note, _validate only works because it is called IMMEDIATELY
                # in the loop, watch out for confusing behavior if it is called
                # later when 'i' has changed
                resulti = resultList[i] = IInputValidator(sub).validate(context, boundTo, data, autoConfigure = False)
                # Merge the valid value in case another fails
                results.update(resulti)
            def _except(e):
                errors = IFormErrors(context.locate(inevow.IRequest).getSession())
                errors.updateErrors(context.key, e.errors)
                pf = e.partialForm
                err = e.errors
                msg = e.formErrorMessage
                ## Set an error message for this group of bindings
                errors.setError(context.key, msg)
                # Merge the failed value
                results.update(pf)
                # Merge the error message
                failures.update(e.errors)
            maybe = exceptblock(_validate, _except, formless.ValidateError)
            if isinstance(maybe, Deferred):
                waiters.append(maybe)
        def _finish(ignored):
            if not failures:
                for specobj, result in zip(spec, resultList):
                    specobj.configure(boundTo, result)
            else:
                #print "There were validation errors. The errors were: ", failures
                raise formless.ValidateError(failures, 'Error:', results)
        return DeferredList(waiters).addBoth(_finish)
    

class MethodBindingValidator(compy.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data, autoConfigure = True):
        """Knows how to validate a dictionary of lists
        where the dictionary may contain a key with the same
        name as some of the arguments to the MethodBinding
        instance.
        """
        typedValue = self.original.typedValue
        results = {}
        failures = {}
        if data.has_key('----'):
            ## ---- is the "direct object", the one argument you can specify using the command line without saying what the argument name is
            data[typedValue.arguments[0].name] = data['----']
            del data['----']
        for binding in typedValue.arguments:
            name = binding.name
            try:
                context = context.with(faketag)
                context.remember(binding, formless.IBinding)
                results[name] = IInputValidator(binding.typedValue).validate(context, boundTo, data.get(name, ['']))
            except formless.InputError, e:
                results[name] = data.get(name, [''])[0]
                failures[name] = e.reason

        if failures:
            #print "There were validation errors. The errors were: ", failures
            raise formless.ValidateError(failures, "Error:", results)

        if autoConfigure:
            def _except(e):
                failures[''] = e.reason # self.original.name
                raise formless.ValidateError(failures, e.reason, results)
            return exceptblock(self.original.configure, _except, formless.InputError,
                               boundTo, results)
        return results


class PropertyBindingValidator(compy.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data, autoConfigure = True):
        """Knows how to validate a dictionary of lists
        where the dictionary may contain a key with the
        same name as the property binding's name.
        """
        binding = self.original
        context.remember(binding, formless.IBinding)
        result = {}
        try:
            result[binding.name] = IInputValidator(binding.typedValue).validate(context, boundTo, data.get(binding.name, ['']))
        except formless.InputError, e:
            result[binding.name] = data.get(binding.name, [''])
            raise formless.ValidateError({binding.name: e.reason}, e.reason, result)

        if autoConfigure:
            return self.original.configure(boundTo, result)
        return result


class TypedValidator(compy.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data):
        """data is a list of strings at this point
        """
        typed = self.original
        if hasattr(typed, 'coerceWithBinding'):
            return typed.coerceWithBinding(data[0], boundTo)
        return typed.coerce(data[0])


class PasswordValidator(compy.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data):
        """Password needs to look at two passwords in the data,
        """
        pw1 = data[0]
        args = context.locate(inevow.IRequest).args
        binding = context.locate(formless.IBinding)
        pw2 = args.get("%s____2" % binding.name, [''])[0]
        if pw1 != pw2:
            raise formless.InputError("Passwords do not match. Please reenter.")
        return self.original.coerce(data[0])


class ListValidator(compy.Adapter):
    def validate(self, context, boundTo, data):
        request = context.locate(inevow.IRequest)
        actee = request.args['freeform-actee']
        subsetIdentifiers = [
            x for x in request.args['freeform-action-selection'][0].split('-') if x
        ]

        subset = []

        binding = context.locate(formless.IConfigurable).getBinding(context, actee)
        fullList = getattr(boundTo, actee)

        for item in fullList:
            identifier = locateConfigurable(context, item).postLocation
            if identifier in subsetIdentifiers:
                subset.append(item)
                subsetIdentifiers.remove(identifier)
                if not subsetIdentifiers:
                    break
        ## TODO: Hack for the pool browser; give the method the store ids and
        ## not the actual item references
        if subsetIdentifiers and not subset:
            return subsetIdentifiers

        return subset


class RequestValidator(compy.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data):
        return context.locate(inevow.IRequest)


class UploadValidator(compy.Adapter):
    __implements__ = IInputValidator,
    def validate(self, context, boundTo, data):
        bind = context.locate(formless.IBinding)
        # TOTAL HACK: this comes from outer space
        fields = context.locate(inevow.IRequest).fields
        try:
            field = fields[bind.name]
            return self.original.coerce(field)
        except KeyError:
            return ''


from nevow import renderer


def instanceQual(instance):
    return util.qual(instance.__class__)


class DefaultTypedInterfaceToIResource(renderer.Renderer):
    document = html[
        head[
            title["Configure an object"],
            link(rel="stylesheet", href="/freeform_css")
        ],
        body[
            "Configure an instance of ", instanceQual, ":",
            p[ configure ]
        ]
    ]
