# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


from nevow import compy


class IInputValidator(compy.Interface):
    """handle a post for a given binding
    """
    def validate(self, context, boundTo, data):
        """do something to boundTo in response to some data

        return a status message if everything's ok, and raise a
        formless.ValidateError if there is a problem
        """
    ## TODO should probably make a distinction between a Typed validator
    ## and a Binding validator; probably would make the code cleaner


class IConfigurableFactory(compy.Interface):
    """A configurable factory is used to find and/or create configurables.
    """
    def locateConfigurable(self, context, name):
        """Return the configurable that responds to the name.
        """


class IResource(compy.Interface):
    def locateChild(self, request, segments):
        """Locate another object which can be adapted to IResource
        Return a tuple of resource, path segments
        """

    def renderHTTP(self, request):
        """Render a request
        """


class IRendererFactory(compy.Interface):
    """A renderer factory is capable of taking a renderer directive (a string)
    and returning a callable which when called, will render a portion of DOM.
    """
    def renderer(self, context, name):
        """Given a context object and a name, return a callable which responds
        to the signature (context, data) or (data) and returns an object which
        is adaptable to ISerializable.
        """


class IData(compy.Interface):
    """Any python object to be used as model data to be passed
    to view functions. Used for marking the context stack only.
    
    ANY python object is said to implement IData.
    """


class IGettable(compy.Interface):
    def get(self, context):
        """Return the data
        
        Return any object
        """


class ISettable(compy.Interface):
    def set(self, context, data):
        """Set the data

        This might be removed.
        """


class IContainer(compy.Interface):
    def child(self, context, name):
        """Return a conceptual child; an attribute, or a key,
        or the result of a function call.
    
        Returns any object; the result may be adapted to IGettable
        if possible.
        
        Return None if the adaptee does not have a child with the
        given name.
        
        TODO: Maybe returning None is bad, and .child should just
        raise whatever exception is natural
        """


class IRequest(compy.Interface):
    """A web request
    """

class ISerializable(compy.Interface):
    ## fixme -- how to describe new interface?
    def serialize(self, context):
        """Serialize the adaptee, with the given context
        stack if necessary.
        """

class IStatusMessage(compy.Interface):
    """A marker interface, which should be set on the user's web session
    to an object which can be cast to a string, which will be shown to the
    user to indicate the status of the most recent operation.
    """


class IHand(compy.Interface):
    """A marker interface which indicates what object the user is currently
    holding in their hand. This is conceptually the "result" of the last operation;
    this interface can be used to mark a status string which indicates whether
    the most recent operation completed successfully, or can be used to hold
    an object which resulted from the most recent operation.
    """


class INewResource(compy.Interface):
    def locateChild(self, request, segments):
        """Given the request and a sequence of strings indictating path segments,
        return a tuple of (rv, segs) where rv is a Resource and segs is a sequence
        of strings which will be passed to rv.locateChild.
        """

    def renderHTTP(self, request):
        """Return a string or a deferred which will fire a string. This string
        will be written to the web browser which initiated this request.
        """

    def willHandle_notFound(self, request):
        """Return a boolean indicating whether this resource wishes to handle a "not found"
        condition for the current request.
        """

    def renderHTTP_notFound(self, request):
        """If this object returned True in response to willHandle_notFound, this
        method will be called to render the Not Found page.
        """


class ICanHandleException(compy.Interface):
    def renderHTTP_exception(self, request, failure):
        """Render an exception to the given request object.
        """


class ICanHandleNotFound(compy.Interface):
    def renderHTTP_notFound(self, request):
        """Render a not found message to the given request.
        """


class IEventMaster(compy.Interface):
    pass


class IDocFactory(compy.Interface):
    def load(self):
        """Return a stan document tree"""
