from cStringIO import StringIO
import os.path

from twisted.python import log
from twisted.internet import defer
from twisted.web import microdom
from twisted.web import resource
from twisted.web import server
from twisted.web.util import redirectTo

from nevow.context import WovenContext, NodeNotFound
from nevow import compy
from nevow import formless
from nevow import inevow
from nevow import liveevil
from nevow import tags


# TODO: Eliminate the following imports, probably by simply pulling them into this module.
# Put these in other modules and make them twisted-free. FourOhFour should be a singleton marker
# called NotFound that doesn't actually do anything.
from nevow.renderer import flatten
from nevow.renderer import FourOhFour
from nevow.renderer import serialize
from nevow.renderer import render
from nevow.renderer import RenderFactory
from nevow.renderer import DataFactory



# TODO:
#
# Remove dependence on t.w.resource.IResource interface
# Add childFactory?
# Allow doc factories to be rendered, for instance:
#     def render_foo(c, d): return span(data=person, render=htmlstr(...))
# More unit tests:
#   * URL traversal


NotFound = FourOhFour(), ()


class LiveEvilChildMixin:
    """Mixin that provides the LiveEvil child resources."""
    child_nevow_liveOutput = liveevil.liveOutput
    child_nevow_liveInput = liveevil.liveInput


class FreeformChildMixin:
    """Mixin that handles locateChild for freeform segments."""
    def locateChild(self, request, segments):
        ## The method or property name we are going to validate against/affect
        bindingName = None

        name = segments[0]
        if name.startswith('freeform_post!'):
            configurableName, bindingName = name.split('!')[1:3]
        elif name.startswith('freeform-action-post!'):
            configurableName, request.args['freeform-actee'] = name.split('!')[1:3]
            bindingName = request.args['freeform-action'][0]
             
        if bindingName:
            ctx = self.context
            if ctx is None:
                ctx = request.context
                self.context = ctx
            ctx.remember(self, inevow.IResource)
            ctx.remember(request, inevow.IRequest)
            from nevow import freeform
            cf = inevow.IConfigurableFactory(self)
            c = cf.locateConfigurable(ctx, configurableName)
            print "locatedConfigurable"
            return self.webFormPost(request, self, c, ctx, bindingName, request.args)
        return FourOhFour(), ()


class ConfigurableFactory:
    """Locates configurables by looking for methods that start with
    configurable_ and end with the name of the configurable. The method
    should take a single arg (other than self) - the current context.
    """
    __implements__ = inevow.IConfigurableFactory

    def locateConfigurable(self, context, name):
        return formless.IConfigurable(getattr(self, 'configurable_%s'%name)(context))

    def configurable_(self, context):
        if compy.implements(self, formless.IConfigurable):
            return self
        if self.original:
            if compy.implements(self.original, formless.TypedInterface):
                return self.original
        return self


class Page(FreeformChildMixin, LiveEvilChildMixin, RenderFactory, DataFactory, ConfigurableFactory):
    """A page is the main Nevow resource and renders a document loaded
    via the document factory (docFactory). A page can also be used as a
    component that is embedded in another page.
    """

    __implements__ = RenderFactory.__implements__, inevow.IResource, DataFactory.__implements__, ConfigurableFactory.__implements__

    original = None
    docFactory = None
    buffered = False

    beforeRender = None
    afterRender = None
    addSlash = None
    
    _context = None

    def __init__(self, original=None, docFactory=None):
        if original:
            self.original = original
        if docFactory:
            self.docFactory = docFactory
        self.toremember = []
        self.children = {}
        self.remember(self, inevow.IConfigurableFactory)

    def putChild(self, name, child):
        self.children[name] = child

    def getParentContext(self, request):
        contexts = getattr(request, 'contexts', [])
        if contexts:
            return contexts[-1]
        return None

    ##
    # IResource methods
    ##

    def locateChild(self, request, segments):

        r = self.children.get(segments[0], None)
        if r:
            return r, segments[1:]

        w = getattr(self, 'child_%s'%segments[0], None)
        if w:
            if compy.implements(w, inevow.IResource) or compy.implements(w, resource.IResource):
                return w, segments[1:]
            return w(request), segments[1:]

        r = self.getDynamicChild(segments[0], request)
        if r:
            return r, segments[1:]

        return FreeformChildMixin.locateChild(self, request, segments)

    def getDynamicChild(self, segment, request):
        return None

    def renderHTTP(self, request):
        from twisted.internet.defer import Deferred
        finished = Deferred()
        if self.beforeRender is not None:
            self.beforeRender(request)
        if self.addSlash and request.prepath[-1] != '':
            return redirectTo(request.URLPath().child(''), request)
            
        log.msg(http_render=None, uri=request.uri)
        
        self.remember(request, inevow.IRequest)
        self.remember(self, inevow.IResource)
        self.remember(self, inevow.IRendererFactory)

        if self.original is None:
            self.remember(self, inevow.IData)
        else:
            self.remember(self.original, inevow.IData)
            self.remember(self.original)

        def finishRequest():
            from nevow import freeform
            sess = request.getSession()
            freeform.IFormErrors(sess).clearAll()
            freeform.IFormDefaults(sess).clearAll()
            self.context = None
            if self.afterRender is not None:
                self.afterRender(request)

        if self.buffered:
            io = StringIO()
            writer = io.write
            def finisher():
                request.write(io.getvalue())
                finishRequest()
                finished.callback('')
        else:
            writer = request.write
            def finisher():
                finishRequest()
                finished.callback('')
            
        if self.context is None:
            ## We were created outside of the normal locateChild lookup
            ## loop, and thus don't have any context to which we are attached.
            ## create one.
            self.context = WovenContext(tag=self)
        assert self.context is not None
        
        render(
            serialize(self.docFactory.load(), self.context),
            self.context, writer, finisher)
        
        return finished

    def renderString(self):
        from twisted.internet.defer import Deferred
        finished = Deferred()
        io = StringIO()
        writer = io.write

        def finisher():
            finished.callback(io.getvalue())
            
        self.context = WovenContext(tag=self)

        render(serialize(self.docFactory.load(), self.context),
               self.context, writer, finisher)

        return finished

    def child_(self, request):
        return self

    def webFormPost(self, request, res, configurable, ctx, bindingName, args):
        def redirectAfterPost():
            print "redirectAfterPost"
            from nevow import freeform
            redirectAfterPost = request.getComponent(freeform.IRedirectAfterPost, None)
            if redirectAfterPost is None:
                ref = request.getHeader('referer') or ''
                from nevow import url
                refpath = url.URLPath.fromString(ref)
                request.redirect(str(refpath))
            else:
                ## Use the redirectAfterPost url
                request.redirect(str(redirectAfterPost))
            from twisted.web import static
            return static.Data('You posted a form to %s' % bindingName, 'text/plain'), ()
        print "webFormPost"
        return defer.maybeDeferred(
            configurable.postForm, ctx, bindingName, args
        ).addCallback(
            self.onPostSuccess, request, ctx, bindingName, redirectAfterPost
        ).addErrback(
            self.onPostFailure, request, ctx, bindingName, redirectAfterPost
        )

    def onPostSuccess(self, result, request, ctx, bindingName, redirectAfterPost):
        print "onPostSuccess"
        sess = request.getSession()
        sess.setComponent(inevow.IHand, result)
        sess.setComponent(inevow.IStatusMessage, "Operation '%s' was completed successfully." % bindingName)
        return redirectAfterPost()

    def onPostFailure(self, reason, request, ctx, bindingName, redirectAfterPost):
        print "onPostFailure"
        from nevow import formless
        reason.trap(formless.ValidateError)
        e = reason.value
        from nevow import freeform
        sess = request.getSession()
        errors = freeform.IFormErrors(sess)
        ## Set the overall error for this form
        errors.setError(ctx.key, e.formErrorMessage)
        errors.updateErrors(bindingName, e.errors)
        K = ctx.key
        if K is None: K = ''
        freeform.IFormDefaults(sess).getAllDefaults(K).update(e.partialForm)
        return redirectAfterPost()

    def remember(self, obj, inter=None):
        """Remember the given object for the given interfaces (or all interfaces
        obj.__implements__) in the top-level context for this page.

        Note that the objects remembered using this API are only remembered
        for the duration of one render.
        """
        if self.context is None:
            self.toremember.append((obj, inter))
        else:
            ## We have a context. Remember this immediately.
            self.context.remember(obj, inter)

    def _setContext(self, new):
        self._context = new
        if new is not None:
            for rememberit in self.toremember:
                new.remember(*rememberit)

    context = property(
        lambda self: self._context,
        _setContext)

    def __call__(self, context, data):
        context.remember(self, inevow.IRendererFactory)
        context.remember(self.original or self, inevow.IData)

        # This tidbit is to enable us to include Page objects inside
        # stan expressions and render_* methods and the like. But
        # because of the way objects can get intertwined, we shouldn't
        # leave the pattern changed.
        old = self.docFactory.pattern
        self.docFactory.pattern = 'content'
        self.docFactory.precompiledDoc = None
        try:
            doc = self.docFactory.load()
            self.docFactory.pattern = old
            self.docFactory.precompiledDoc = None
        except NodeNotFound:
            self.docFactory.pattern = old
            self.docFactory.precompiledDoc = None
            doc = self.docFactory.load()
        return doc
    
class DocFactory:
    """Base class for document factories. I am responsible for loading a document template for
    the renderer to use."""

    __implements__ = inevow.IDocFactory,
    
    pattern = None
    precompiledDoc = None

    def __init__(self, pattern=None):
        if pattern:
            self.pattern = pattern

    def precompile(self):
        context = WovenContext(precompile=True)
        doc = flatten(serialize(self.getDoc(), context))
        if self.pattern:
            ctx = context.with(tags.invisible[doc])
            doc = [ctx.onePattern(self.pattern)]
        return doc

    def load(self):
        """Load and return the document for the renderer"""
        if self.precompiledDoc is None:
            self.precompiledDoc = self.precompile()
        return self.precompiledDoc


# Caches for the document factories
_stanCache = {}
_htmlstrCache = {}
_htmlfileCache = {}
_xmlstrCache = {}
_xmlfileCache = {}


class stan(DocFactory):
    """A stan tags document factory"""

    def __init__(self, doc, **kw):
        self.doc = doc
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        return self.doc


class htmlstr(DocFactory):
    """A document factory for HTML contained in a string"""

    beExtremelyLenient = False

    def __init__(self, template, beExtremelyLenient=None, **kw):
        self.template = template
        if beExtremelyLenient:
            self.beExtremelyLenient = beExtremelyLenient
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        return microdom.parseString(self.template, beExtremelyLenient=self.beExtremelyLenient)

    def precompile(self):
        doc = _htmlstrCache.get((self.template, self.pattern), None)
        if doc:
            return doc
        doc = DocFactory.precompile(self)
        _htmlstrCache[(self.template, self.pattern)] = doc
        return doc


class htmlfile(DocFactory):
    """A document factory for an HTML disk template"""

    template = None
    templateDir = ''
    beExtremelyLenient = False
    _mtime = None

    def __init__(self, template, templateDir=None, beExtremelyLenient=None, **kw):
        self.template = template
        if templateDir is not None:
            self.templateDir = templateDir
        if beExtremelyLenient is not None:
            self.beExtremelyLenient = beExtremelyLenient
        self._filename = os.path.join(self.templateDir, self.template)
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        return microdom.parse(self._filename, beExtremelyLenient=self.beExtremelyLenient)

    def precompile(self):
        mtime, doc = _htmlfileCache.get((self._filename, self.pattern), (None,None))
        self._mtime = os.path.getmtime(self._filename)
        if mtime == self._mtime:
            return doc
        doc = DocFactory.precompile(self)
        _htmlfileCache[(self._filename, self.pattern)] = self._mtime, doc
        return doc
        
    def load(self):
        if self._mtime != os.path.getmtime(self._filename):
            self.precompiledDoc = None
        return DocFactory.load(self)


class xmlstr(DocFactory):

    doc = None

    def __init__(self, template, **kw):
        self.template = template
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        from nevow.flat import flatsax
        return flatsax.parseString(self.template)

    def precompile(self):
        doc = _xmlstrCache.get((self.template, self.pattern), None)
        if doc:
            return doc
        doc = DocFactory.precompile(self)
        _xmlstrCache[(self.template, self.pattern)] = doc
        return doc
        

class xmlfile(DocFactory):

    template = None
    templateDir = ''

    _mtime = None

    def __init__(self, template, templateDir=None, **kw):
        self.template = template
        if templateDir:
            self.templateDir = templateDir
        self._filename = os.path.join(self.templateDir, self.template)
        DocFactory.__init__(self, **kw)

    def getDoc(self):
        from nevow.flat import flatsax
        return flatsax.parse(file(self._filename))
    
    def precompile(self):
        mtime, doc = _xmlfileCache.get((self._filename, self.pattern), (None,None))
        self._mtime = os.path.getmtime(self._filename)
        if mtime == self._mtime:
            return doc
        doc = DocFactory.precompile(self)
        _xmlfileCache[(self._filename, self.pattern)] = self._mtime, doc
        return doc
        
    def load(self):
        if self._mtime != os.path.getmtime(self._filename):
            self.precompiledDoc = None
        return DocFactory.load(self)


def DocFactorySerializer(original, context):
    return serialize(original.load(), context)
