from __future__ import generators
import weakref

from nevow import inevow
from nevow.stan import xml


import urlparse
import urllib

class URLPath:
    keepQuery = False
    def __init__(self, scheme='', netloc='localhost', path='',
                 query='', fragment=''):
        self.scheme = scheme or 'http'
        self.netloc = netloc
        self.path = path or '/'
        self.query = query
        self.fragment = fragment

    _qpathlist = None
    _uqpathlist = None
    
    def pathList(self, unquote=0, copy=1):
        if self._qpathlist is None:
            self._qpathlist = self.path.split('/')
            self._uqpathlist = map(urllib.unquote, self._qpathlist)
        if unquote:
            result = self._uqpathlist
        else:
            result = self._qpathlist
        if copy:
            return result[:]
        else:
            return result

    def fromString(klass, st):
        t = urlparse.urlsplit(st)
        u = klass(*t)
        return u

    fromString = classmethod(fromString)

    def fromRequest(klass, request):
        R = klass.fromString(request.prePathURL())
        if '?' in request.uri:
            R.query = request.uri.split('?')[-1]
        return R

    fromRequest = classmethod(fromRequest)

    def _pathMod(self, newpathsegs, keepQuery):
        if keepQuery:
            query = self.query
        else:
            query = ''
        return URLPath(self.scheme,
                        self.netloc,
                        '/'.join(newpathsegs),
                        query)

    def sibling(self, path, keepQuery=0):
        l = self.pathList()
        l[-1] = path
        return self._pathMod(l, keepQuery)

    def child(self, path, keepQuery=0):
        l = self.pathList()
        if l[-1] == '':
            l[-1] = path
        else:
            l.append(path)
        return self._pathMod(l, keepQuery)

    def parent(self, keepQuery=0):
        l = self.pathList()
        if l[-1] == '':
            del l[-2]
        else:
            # We are a file, such as http://example.com/foo/bar
            # our parent directory is http://example.com/
            l.pop()
            l[-1] = ''
        return self._pathMod(l, keepQuery)

    def here(self, keepQuery=0):
        l = self.pathList()
        if l[-1] != '':
            l[-1] = ''
        return self._pathMod(l, keepQuery)

    def click(self, st):
        """Return a path which is the URL where a browser would presumably take
        you if you clicked on a link with an HREF as given.
        """
        scheme, netloc, path, query, fragment = urlparse.urlsplit(st)
        if not scheme:
            scheme = self.scheme
        if not netloc:
            netloc = self.netloc
            if not path:
                path = self.path
                if not query:
                    query = self.query
            elif path[0] != '/':
                l = self.pathList()
                l[-1] = path
                path = '/'.join(l)
        
        return URLPath(scheme,
                        netloc,
                        path,
                        query,
                        fragment)


    
    def __str__(self):
        x = urlparse.urlunsplit((
            self.scheme, self.netloc, self.path,
            self.query, self.fragment))
        return x

    def __repr__(self):
        return ('URLPath(scheme=%r, netloc=%r, path=%r, query=%r, fragment=%r)'
                % (self.scheme, self.netloc, self.path, self.query, self.fragment))


class URLOverlay(object):
    def __init__(self, urlaccessor, appendquery=None, dolater=None):
        """URLOverlay -- a Proto like object for abstractly specifying urls in stan trees

        @ivar urlaccessor: a function which takes context and returns a URLPath
        @ivar appendquery: a list of tuples of (key, value) to append to the query string
        @ivar dolater: a list of tuples of (command, args, kw) where args is a tuple and kw
            is a dict; when the URLPath is returned from urlaccessor during rendering, these
            methods will be applied to the URLPath in order
        """
        self.urlaccessor = urlaccessor
        if appendquery is None:
            self.appendquery = []
        else:
            self.appendquery = appendquery
        if dolater is None:
            self.dolater = []
        else:
            self.dolater = dolater

    def __call__(self, *args, **kw):
        q = self.appendquery[:]
        q.extend(args)
        q.extend(kw.items())
        return self.__class__(self.urlaccessor, q, self.dolater)

    def addCommand(self, cmd, args, kw):
        dl = self.dolater[:]
        dl.append((cmd, args, kw))
        return self.__class__(self.urlaccessor, self.appendquery, dl)


def createForwarder(cmd):
    return lambda self, *args, **kw: self.addCommand(cmd, args, kw)


for cmd in ['sibling', 'child', 'parent', 'here', 'click']:
    setattr(URLOverlay, cmd, createForwarder(cmd))


def hereaccessor(context):
    req = context.locate(inevow.IRequest)
    return req.URLPath()


here = URLOverlay(hereaccessor)


def gethereaccessor(context):
    req = context.locate(inevow.IRequest)
    rup = req.URLPath()
    rup.keepQuery = True
    return rup

gethere = URLOverlay(gethereaccessor)


def rootaccessor(context):
    req = context.locate(inevow.IRequest)
    root = req.getRootURL()
    if root is None:
        return URLPath.fromString(req.prePathURL()).click('/')
    return URLPath.fromString(root)


root = URLOverlay(rootaccessor)


def URLOverlaySerializer(original, context):
    if context.precompile:
        yield original
    else:
        url = original.urlaccessor(context)
        aqk = [k for (k, v) in original.appendquery]
        for (cmd, args, kw) in original.dolater:
            url = getattr(url, cmd)(*args, **kw)
        if url.keepQuery:
            if url.query:
                qu = [
                    (k, v) for (k, v) in
                        [tuple(x.split('=')) for x in url.query.split('&')]
                    if k not in aqk
                ]
            else: qu = []
        else:
            qu = []
        qu.extend(original.appendquery)
        url.query = urllib.urlencode(qu)
        yield xml(str(url))


## This is totally unfinished and doesn't work yet.
#class IURLGenerator(compy.Interface):
#    pass


class URLGenerator:
    #__implements__ = IURLGenerator,

    def __init__(self):
        self._objmap = weakref.WeakKeyDictionary()

    def objectMountedAt(self, obj, at):
        self._objmap[obj] = at

    def url(self, obj):
        try:
            return self._objmap.get(obj, None)
        except TypeError:
            return None

    __call__ = url

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_objmap']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._objmap = weakref.WeakKeyDictionary()


